// @@@@@@@@@@ Original Code by bitluni
// @@@@@@@@@@ From https://github.com/bitluni/ESP32CameraI2S

// This code is for ESP32, ST7735 and OV7670 without FIFO
// ESP32 that we used was NodeMCU-32s
// Data sheet link http://wiki.ai-thinker.com/_media/esp32/docs/nodemcu-32s_product_specification.pdf
// ST7735 purchase link https://www.arduitronics.com/product/187/%E0%B8%88%E0%B8%AD-tft-lcd-%E0%B8%82%E0%B8%99%E0%B8%B2%E0%B8%94-1-8-%E0%B8%9E%E0%B8%A3%E0%B9%89%E0%B8%AD%E0%B8%A1%E0%B8%8A%E0%B9%88%E0%B8%AD%E0%B8%87%E0%B9%80%E0%B8%AA%E0%B8%B5%E0%B8%A2%E0%B8%9A-sd-card
// OV7670 without FIFO purchase link https://www.arduitronics.com/product/120/camera-module-ov7670

// Libraries
// Arduino core for the ESP32 https://github.com/espressif/arduino-esp32#installation-instructions
// Adafruit GFX Library
// Adafruit ST7735 Library
// ESPAsyncWebServer Library https://github.com/me-no-dev/ESPAsyncWebServer
// Async TCP Library https://github.com/me-no-dev/AsyncTCP

#include "OV7670.h"

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library

#include <WiFi.h>
#include <HTTPClient.h>
#include "ESPAsyncWebServer.h" 

const char* ssid = "Wifi Name"; 
const char* password =  "Wifi Password";

volatile int state = 1; // State 1 = Open camera, State 0 = Close camera

IPAddress local_IP(172, 20, 10, 4);   // Static IP for your ESP32
IPAddress gateway(172, 20, 10, 1);    // Gateway IP
IPAddress subnet(255, 255, 255, 240); // Subnetmask

AsyncWebServer server(8090);          // Listen on port 8090

const int SIOD = 21; //SDA
const int SIOC = 22; //SCL

const int VSYNC = 34;
const int HREF = 35;

const int XCLK = 32;
const int PCLK = 33;

const int D0 = 27;
const int D1 = 17;
const int D2 = 16;
const int D3 = 15;
const int D4 = 14;
const int D5 = 13;
const int D6 = 12;
const int D7 = 4;

const int TFT_DC = 2;
const int TFT_CS = 5;
//DIN <- MOSI 23
//CLK <- SCK 18

Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS,  TFT_DC, 0/*no reset*/);
OV7670 *camera;

void displayY8(unsigned char * frame, int xres, int yres) // I didn't use this
{
  tft.setAddrWindow(0, 0, yres - 1, xres - 1);
  int i = 0;
  for(int x = 0; x < xres; x++)
    for(int y = 0; y < yres; y++)
    {
      i = y * xres + x;
      unsigned char c = frame[i];
      unsigned short r = c >> 3;
      unsigned short g = c >> 2;
      unsigned short b = c >> 3;
      tft.pushColor(r << 11 | g << 5 | b);
    }  
}

void displayRGB565(unsigned char * frame, int xres, int yres)
{
  tft.setAddrWindow(0, 0, yres-1, xres-1);
  int i = 0;
  for(int y = 0; y < yres+8; y++){ // +8 as LCD(ST7735)'s resolution is 160x128 not 160x120
    for(int x = 0; x < xres; x++)
    {
      i = (y * xres + x) << 1;
      tft.pushColor((frame[i] | (frame[i+1] << 8)));
    }
  }  
}

void setup() 
{
  Serial.begin(115200);
  camera = new OV7670(OV7670::Mode::QQVGA_RGB565, SIOD, SIOC, VSYNC, HREF, XCLK, PCLK, D0, D1, D2, D3, D4, D5, D6, D7);
  
  tft.initR(INITR_BLACKTAB);
  tft.setRotation(1);
  tft.fillScreen(0);
  camera->oneFrame(); // oneFrame() is a method to capture a frame from OV7670
  displayRGB565(camera->frame, camera->xres, camera->yres); // Display on LCD 
  
  //wifi
  WiFi.config(local_IP, gateway, subnet); 
 
  WiFi.begin(ssid, password);
 
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }
 
  Serial.println("Connected to the WiFi network");
  Serial.println(WiFi.localIP());

  // @@@@@@@@@@ Example Code by Rui Santos
  // @@@@@@@@@@ From https://randomnerdtutorials.com/esp32-client-server-wi-fi/
  
  server.on("/frame", HTTP_GET, [](AsyncWebServerRequest *request){ // Waiting for GET from another device that wants an image

    AsyncWebServerResponse *response = request->beginResponse_P(200, "image/bmp", (uint8_t*)camera->frame, camera-> xres * camera-> yres * 2); // Sending as byte stream, camera->frame is unsigned char* so casting to uint8_t* should be needed, Total packet size is 160*120*2, as one pixel is 2 bytes.
    request->send(response);
  });

  server.on("/noti", HTTP_POST, [](AsyncWebServerRequest *request){ // Waiting for POST for command to open or close camera

    request->send(200);
    int params = request->params(); // Expecting data like this >>>data={'mode':1}<<<
    for (int i = 0; i < params; i++) {
      AsyncWebParameter* p = request->getParam(i);

      String para = p->name().c_str();
      String result = p->value().c_str();

      if (result == "0") {
        state = 0; 
      }
      else if (result == "1") {
        state = 1;  
      }
    }
    request->send(200);
    
  });

  server.begin();
}


void loop()
{
  camera->oneFrame(); // Get an image
  //delay(50);
  if (state == 1) {
    displayRGB565(camera->frame, camera->xres, camera->yres); // If state = 1, Display on LCD
  }
  else if (state == 0) { // If state = 0, Display black screen on LCD
    tft.fillScreen(0);
  }
  Serial.println(state); 
}
